#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file DroneController.py
 @brief Control the flight of the drone.
 @date $Date$

 @author East Japan Accounting Center Co., Ltd

 (c)2017 The University of Aizu
 This software is released under the New BSD License.

"""
import sys
import math
import csv
from datetime import datetime

# Import RTM module
import RTC
import OpenRTM_aist

import target_gps
import drone_distance_cal

# Setup logger
from logging import getLogger, StreamHandler
from logging import DEBUG as LOG_LV

sys.path.append(".")

logger = getLogger(__name__)
handler = StreamHandler()
logger.setLevel(LOG_LV)
handler.setLevel(LOG_LV)
logger.addHandler(handler)

# constants
BUTTON_01_MASK = 0x0001
BUTTON_02_MASK = 0x0002
BUTTON_03_MASK = 0x0004
BUTTON_04_MASK = 0x0008
BUTTON_05_MASK = 0x0010
BUTTON_06_MASK = 0x0020
BUTTON_07_MASK = 0x0040
BUTTON_08_MASK = 0x0080
BUTTON_09_MASK = 0x0100
BUTTON_10_MASK = 0x0200
BUTTON_11_MASK = 0x0400
BUTTON_12_MASK = 0x0800
BUTTON_13_MASK = 0x1000
BUTTON_14_MASK = 0x2000

MOVE_SPEED_MAX = 10.0
MOVE_SPEED_MIN = 1.0

MOVE_UPDATE_COEF_ABSOLUTE_NS = 1.0
MOVE_UPDATE_COEF_ABSOLUTE_EW = 0.65
MOVE_UPDATE_COEF_RELATIVE = 0.55

ALLOWABLE_ERROR_POS_CMP = 5.0

ALLOWABLE_ERROR_HEIGHT = 0.0

MODE_ABSOLUTE = 1

ALT_CHG_CANCEL_WAIT_TM = 50  # 5sec(=exec rate(10Hz))

# file mode
FILE_MODE_WRITE = "w"
FILE_MODE_ADD_WRITE = "a"

# Import Service implementation class
# <rtc-template block="service_impl">

# </rtc-template>

# Import Service stub modules
# <rtc-template block="consumer_import">
# </rtc-template>


# This module's spesification
# <rtc-template block="module_spec">
dronecontroller_spec = ["implementation_id", "DroneController",
         "type_name",         "DroneController",
         "description",       "Control the flight of the drone.",
         "version",           "1.4.0",
         "vendor",            "EAC",
         "category",          "Drone",
         "activity_type",     "STATIC",
         "max_instance",      "1",
         "language",          "Python",
         "lang_type",         "SCRIPT",
         "conf.default.AltCoefficient", "1",
         "conf.default.LatLonCoefficient", "10",
         "conf.default.LatDiff10m", "0.000137",
         "conf.default.LonDiff10m", "0.000108",
         "conf.default.MoveSpeed", "3.0",
         "conf.default.MinAlt", "2.0",
         "conf.default.MaxAlt", "15.0",
         "conf.default.TurnSpeed", "45",
         "conf.default.NoseLatDiff10m", "9.01323",
         "conf.default.NoseLonDiff10m", "8.9831566",
         "conf.default.MoveTurnSpeed", "30",
         "conf.__widget__.AltCoefficient", "spin",
         "conf.__widget__.LatLonCoefficient", "spin",
         "conf.__widget__.LatDiff10m", "spin",
         "conf.__widget__.LonDiff10m", "spin",
         "conf.__widget__.MoveSpeed", "spin",
         "conf.__widget__.MinAlt", "spin",
         "conf.__widget__.MaxAlt", "spin",
         "conf.__widget__.TurnSpeed", "spin",
         "conf.__widget__.NoseLatDiff10m", "spin",
         "conf.__widget__.NoseLonDiff10m", "spin",
         "conf.__widget__.MoveTurnSpeed", "spin",
         "conf.__constraints__.AltCoefficient", "1<=x<=10",
         "conf.__constraints__.LatLonCoefficient", "2<=x<=10",
         "conf.__constraints__.LatDiff10m", "0.000000<x<0.100000",
         "conf.__constraints__.LonDiff10m", "0.000000<x<0.100000",
         "conf.__constraints__.MoveSpeed", "1.0<=x<=10.0",
         "conf.__constraints__.MinAlt", "1.0<=x<=100.0",
         "conf.__constraints__.MaxAlt", "10.0<=x<=100.0",
         "conf.__constraints__.TurnSpeed", "0<=x<=360",
         "conf.__constraints__.NoseLatDiff10m", "0.000000<=x<=10.000000",
         "conf.__constraints__.NoseLonDiff10m", "0.0000000<=x<=10.0000000",
         "conf.__constraints__.MoveTurnSpeed", "0<=x<=360",
         ""]
# </rtc-template>

##
# @class DroneController
# @brief Control the flight of the drone.
#
# Refer to the "機能仕様書_DroneController" for details.
#
# Refer to the "機能仕様書_DroneController" for details.
#
# Refer to the "機能仕様書_DroneController" for details.
#
# Refer to the "機能仕様書_DroneController" for details.
#
#


class DroneController(OpenRTM_aist.DataFlowComponentBase):

    ##
    # @brief constructor
    # @param manager Maneger Object
    #
    def __init__(self, manager):
        OpenRTM_aist.DataFlowComponentBase.__init__(self, manager)

        self._d_ButtonIn = RTC.TimedULong(RTC.Time(0, 0), 0)
        """
        PS4コントローラのボタン入力状態
         - Type: RTC::TimedULong
        """
        self._ButtonInIn = OpenRTM_aist.InPort("ButtonIn", self._d_ButtonIn)
        self._d_AnalogIn = RTC.TimedDoubleSeq(RTC.Time(0, 0), [])
        """
        PS4コントローラのアナログ入力状態
         - Type: RTC::TimedDoubleSeq
        """
        self._AnalogInIn = OpenRTM_aist.InPort("AnalogIn", self._d_AnalogIn)
        self._d_inDrCtrl = RTC.DroneCtrl(RTC.Time(0, 0),
                                         0,
                                         RTC.TimedString(RTC.Time(0, 0), ""),
                                         RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                     100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                     0.0, 0, RTC.GPS_FIX_NORMAL),
                                         RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                     100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                     0.0, 0, RTC.GPS_FIX_NORMAL),
                                         RTC.TimedFloatSeq(
                                             RTC.Time(0, 0), [0.0, 0.0, 0.0]),
                                         RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                         RTC.TimedString(RTC.Time(0, 0), ""))
        """
        ドローンコントロール情報
        以下の情報をパッケージ
        ・受信ID(DroneIOからの送信ID)
        ・ホーム(基地局)座標
        ・ドローン現在位置座標
        ・加速度
        ・磁気コンパス
        ・ドローンI/Oの内部ステータス
         - Type: RTC::DroneCtrl
        """
        self._inDrCtrlIn = OpenRTM_aist.InPort("inDrCtrl", self._d_inDrCtrl)
        self._d_outTarget = RTC.DroneTarget(RTC.Time(0, 0),
                                            0,
                                            RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                        100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                        0.0, 0, RTC.GPS_FIX_NORMAL),
                                            RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                            RTC.TimedShortSeq(RTC.Time(0, 0), [0, 0]))
        """
        ドローンターゲット情報
        以下の情報をパッケージ
        ・送信ID(1から始まる連番)
        ・ターゲット座標
        ・ターゲット速度
        ・旋回方向及び移動量
         - Type: RTC::DroneTarget
        """
        self._outTargetOut = OpenRTM_aist.OutPort("outTarget", self._d_outTarget)
        self._d_outTargetSpeed = RTC.TimedFloat(RTC.Time(0, 0), 0)
        """
        ターゲット速度
         - Type: RTC::TimedFloat
        """
        self._outTargetSpeedOut = OpenRTM_aist.OutPort("outTargetSpeed", self._d_outTargetSpeed)
        self._d_outMoveMode = RTC.TimedString(RTC.Time(0, 0), 0)
        """
        移動モード(Absolute or Relative)
         - Type: RTC::TimedString
        """
        self._outMoveModeOut = OpenRTM_aist.OutPort("outMoveMode", self._d_outMoveMode)

        # initialize of configuration-data.
        # <rtc-template block="init_conf_param">
        """
        移動ターゲット高度算出係数
         - Name: AltCoefficient AltCoefficient
         - DefaultValue: 1
         - Range: 1<=x<=10
        """
        self._AltCoefficient = [1]
        """
        移動ターゲット経緯度算出係数
         - Name: LatLonCoefficient LatLonCoefficient
         - DefaultValue: 10
         - Range: 2<=x<=10
        """
        self._LatLonCoefficient = [10]
        """
        10mあたりの緯度差分
         - Name: LatDiff10m LatDiff10m
         - DefaultValue: 0.000137
         - Range: 0.000000<x<0.100000
        """
        self._LatDiff10m = [0.000137]
        """
        10mあたりの経度差分
         - Name: LonDiff10m LonDiff10m
         - DefaultValue: 0.000108
         - Range: 0.000000<x<0.100000
        """
        self._LonDiff10m = [0.000108]
        """
        ドローン移動速度
         - Name: Speed MoveSpeed
         - DefaultValue: 3.0
         - Range: 1.0<=x<=10.0
        """
        self._MoveSpeed = [3.0]
        """
        ドローン最小高度
         - Name: MinAlt MinAlt
         - DefaultValue: 2.0
         - Range: 1.0<=x<=100.0
        """
        self._MinAlt = [2.0]
        """
        ドローン最大高度
        ※ドローン最大高度(MaxAlt)の設定値は、最小高度(MinAlt)の設定値以上の値とする
        。
         - Name: MaxAlt MaxAlt
         - DefaultValue: 15.0
         - Range: 10.0<=x<=100.0
        """
        self._MaxAlt = [15.0]
        """
        ドローン旋回速度
         - Name: TurnSpeed TurnSpeed
         - DefaultValue: 45
         - Range: 0<=x<=360
        """
        self._TurnSpeed = [45]
        """
        10mあたりの機首緯度差分
         - Name: NoseLatDiff10m NoseLatDiff10m
         - DefaultValue: 9.01323
         - Range: 0.000000<=x<=10.000000
        """
        self._NoseLatDiff10m = [9.01323]
        """
        10mあたりの機首経度差分
         - Name: NoseLonDiff10m NoseLonDiff10m
         - DefaultValue: 8.9831566
         - Range: 0.0000000<=x<=10.0000000
        """
        self._NoseLonDiff10m = [8.9831566]
        """
        ドローン移動旋回速度
         - Name: MoveTurnSpeed MoveTurnSpeed
         - DefaultValue: 30
         - Range: 0<=x<=360
        """
        self._MoveTurnSpeed = [30]

        # </rtc-template>

        self._target_gps = target_gps.TargetGPS()

        self._button = False
        self._analog = False
        self._drone_gps = False

        self._SpeedParam = self._MoveSpeed[0]

        self._last_front_back = 0
        self._last_right_left = 0
        self._last_up_down = 0
        self._last_speedparam = 0
        self._last_move_mode = "Absolute"

        self._alt_chg_req_cancel = False
        self._alt_chg_cancel_wait_tm = 0

        self._file_name = ""
        self._file = False
        self._writer = False
        self._header = ["Category", "ActStartDate", "ActStartTime", "RcvDate", "RcvTime", "RcvRobotID", "RcvID",
                       "HomeLat", "HomeLon", "HomeAlt", "RobotLat", "RobotLon", "RobotAlt",
                       "AccX", "AccY", "AccZ", "Mag", "State", "SndDate", "SndTime", "SndRobotID", "SndID", "MoveMode",
                       "TargetLat", "TargetLon", "TargetAlt", "TargetSpd", "TargetYawDir", "TargetYawDeg", "RTCState"]
        self._act_start_date = ""
        self._act_start_time = ""

        self._out_target_wrote = False

        # connection check for data port ButtonInIn
        func = lambda: self._cb_connect_status_change("ButtonIn", True)
        ob = ConnectionCallback(func)
        self._ButtonInIn.setOnConnected(ob)

        func = lambda: self._cb_connect_status_change("ButtonIn", False)
        ob = ConnectionCallback(func)
        self._ButtonInIn.setOnDisconnected(ob)

        # connection check for data port AnalogInIn
        func = lambda: self._cb_connect_status_change("AnalogIn", True)
        ob = ConnectionCallback(func)
        self._AnalogInIn.setOnConnected(ob)

        func = lambda: self._cb_connect_status_change("AnalogIn", False)
        ob = ConnectionCallback(func)
        self._AnalogInIn.setOnDisconnected(ob)

    ##
    # RTC初期化
    #
    # The initialize action (on CREATED->ALIVE transition)
    # formaer rtc_init_entry()
    #
    # @return RTC::ReturnCode_t
    #
    #
    def onInitialize(self):
        # Bind variables and configuration variable
        self.bindParameter("AltCoefficient", self._AltCoefficient, "1")
        self.bindParameter("LatLonCoefficient", self._LatLonCoefficient, "10")
        self.bindParameter("LatDiff10m", self._LatDiff10m, "0.000137")
        self.bindParameter("LonDiff10m", self._LonDiff10m, "0.000108")
        self.bindParameter("MoveSpeed", self._MoveSpeed, "3.0")
        self.bindParameter("MinAlt", self._MinAlt, "2.0")
        self.bindParameter("MaxAlt", self._MaxAlt, "15.0")
        self.bindParameter("TurnSpeed", self._TurnSpeed, "45")
        self.bindParameter("NoseLatDiff10m", self._NoseLatDiff10m, "9.01323")
        self.bindParameter("NoseLonDiff10m", self._NoseLonDiff10m, "8.9831566")
        self.bindParameter("MoveTurnSpeed", self._MoveTurnSpeed, "30")

        # Set InPort buffers
        self.addInPort("ButtonIn", self._ButtonInIn)
        self.addInPort("AnalogIn", self._AnalogInIn)
        self.addInPort("inDrCtrl", self._inDrCtrlIn)

        # Set OutPort buffers
        self.addOutPort("outTarget", self._outTargetOut)
        self.addOutPort("outTargetSpeed", self._outTargetSpeedOut)
        self.addOutPort("outMoveMode", self._outMoveModeOut)

        # Set service provider to Ports

        # Set service consumers to Ports

        # Set CORBA Service Ports

        return RTC.RTC_OK

        ##
        #
        # The finalize action (on ALIVE->END transition)
        # formaer rtc_exiting_entry()
        #
        # @return RTC::ReturnCode_t

        #
    def onFinalize(self):

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()
        self._log_row = ["NULL"] * len(self._header)
        self._write_rtc_viewer_log("Exit")
        self._close_rtc_viewer_log()

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The startup action when ExecutionContext startup
    #    # former rtc_starting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onStartup(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The shutdown action when ExecutionContext stop
    #    # former rtc_stopping_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onShutdown(self, ec_id):
    #
    #    return RTC.RTC_OK

        ##
        # 変数初期化
        #
        # The activated action (Active state entry action)
        # former rtc_active_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onActivated(self, ec_id):

        self._file_name = ""
        self._file = None
        self._writer = None
        self._act_start_date, self._act_start_time = self._get_datetime_to_str()

        init_movemode = MODE_ABSOLUTE
        self._target_gps.set_controller_mode(init_movemode)
        self._SpeedParam = self._MoveSpeed[0]

        self._last_move_mode = "Absolute"

        # viewer-log open and write(header-row)
        self._log_row = ["NULL"] * len(self._header)
        self._ready_to_log_write(FILE_MODE_WRITE)

        return RTC.RTC_OK

        ##
        #
        # The deactivated action (Active state exit action)
        # former rtc_active_exit()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onDeactivated(self, ec_id):

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()
        self._log_row = ["NULL"] * len(self._header)
        self._write_rtc_viewer_log("Inactive")
        self._close_rtc_viewer_log()

        return RTC.RTC_OK

        ##
        # PS4キー入力、ドローンステータス取得、およびコントロールデータ送信の周期処理。
        #
        # The execution action that is invoked periodically
        # former rtc_active_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onExecute(self, ec_id):

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()

        self._read_button()
        self._read_analog()
        self._read_drone_ctrl()

        self._set_target_speed()
        self._set_move_mode()
        self._set_target_yaw()
        self._set_target_gps()
        self._set_outport()

        self._log_row = ["NULL"] * len(self._header)
        self._write_rtc_viewer_log("Active")

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The aborting action when main logic error occurred.
    #    # former rtc_aborting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onAborting(self, ec_id):
    #
    #    return RTC.RTC_OK

        ##
        #
        # The error action in ERROR state
        # former rtc_error_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onError(self, ec_id):

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()

        if self._file is None:
            self._ready_to_log_write(FILE_MODE_ADD_WRITE)

        self._log_row = ["NULL"] * len(self._header)
        self._write_rtc_viewer_log("Error")

        return RTC.RTC_OK

        ##
        #
        # The reset action that is invoked resetting
        # This is same but different the former rtc_init_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onReset(self, ec_id):

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()
        self._log_row = ["NULL"] * len(self._header)
        self._write_rtc_viewer_log("Inactive")
        self._close_rtc_viewer_log()

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The state update action that is invoked after onExecute() action
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #

    #    #
    # def onStateUpdate(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The action that is invoked when execution context's rate is changed
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onRateChanged(self, ec_id):
    #
    #    return RTC.RTC_OK

    def _read_button(self):
        self._d_ButtonIn = RTC.TimedULong(RTC.Time(0, 0), 0)

        if self._button:
            self._d_ButtonIn = self._ButtonInIn.read()
            logger.debug("Button = 0b" + format(self._d_ButtonIn.data, '016b'))
        else:
            logger.error("ERROR: ButtonIn is NOT connected")

    def _read_analog(self):
        self._d_AnalogIn = RTC.TimedDoubleSeq(RTC.Time(0, 0), [])

        if self._analog:
            self._d_AnalogIn = self._AnalogInIn.read()
            for i in range(0, len(self._d_AnalogIn.data), 2):
                d1 = d2 = None
                try:
                    d1 = self._d_AnalogIn.data[i]
                    d2 = self._d_AnalogIn.data[i + 1]
                except IndexError:
                    pass

                if d2 is None:
                    logger.debug("Analog[%d]:  % f", i, d1)
                else:
                    logger.debug("Analog[%d,%d]: % f % f", i, i + 1, d1, d2)
        else:
            logger.error("ERROR: AnalogIn is NOT connected")

    def _read_drone_ctrl(self):
        self._drone_gps = False
        self._d_inDrCtrl = RTC.DroneCtrl(RTC.Time(0,0),
                                         0,
                                         RTC.TimedString(RTC.Time(0, 0), ""),
                                         RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                     100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                     0.0, 0, RTC.GPS_FIX_NORMAL),
                                         RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                     100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                     0.0, 0, RTC.GPS_FIX_NORMAL),
                                         RTC.TimedFloatSeq(RTC.Time(0,0), [0.0, 0.0, 0.0]),
                                         RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                         RTC.TimedString(RTC.Time(0, 0), ""))

        # read_drone_ctrl
        if self._inDrCtrlIn.isNew():
            self._d_inDrCtrl = self._inDrCtrlIn.read()
            OpenRTM_aist.setTimestamp(self._d_inDrCtrl)

            if self._is_valid_drctrl(self._d_inDrCtrl, 0):
                self._drone_gps = True

                logger.debug("Rcv ID:    % d", self._d_inDrCtrl.id)

                logger.debug("RobotID:   %s", self._d_inDrCtrl.robotID.data)

                logger.debug("HomeGPS:   % f % f % f",
                            self._d_inDrCtrl.homeGPS.latitude,
                            self._d_inDrCtrl.homeGPS.longitude,
                            self._d_inDrCtrl.homeGPS.altitude)

                logger.debug("DroneGPS:  % f % f % f",
                            self._d_inDrCtrl.droneGPS.latitude,
                            self._d_inDrCtrl.droneGPS.longitude,
                            self._d_inDrCtrl.droneGPS.altitude)

                logger.debug("Mag:       %d", self._d_inDrCtrl.mag.data)
                logger.debug("State:     %s", self._d_inDrCtrl.state.data)

    def _config_target_gps(self):
        config = {"AltCoefficient": self._AltCoefficient[0],
                  "LatLonCoefficient": self._LatLonCoefficient[0],
                  "LatDiff10m": self._LatDiff10m[0],
                  "LonDiff10m": self._LonDiff10m[0],
                  "MoveSpeed": self._MoveSpeed[0],
                  "MinAlt": self._MinAlt[0],
                  "MaxAlt": self._MaxAlt[0],
                  "TurnSpeed": self._TurnSpeed[0],
                  "NoseLatDiff10m": self._NoseLatDiff10m[0],
                  "NoseLonDiff10m": self._NoseLonDiff10m[0],
                  "MoveTurnSpeed": self._MoveTurnSpeed[0]
                  }

        self._target_gps.config_param(config)

    def _set_target_gps(self):
        if not self._drone_gps:
            logger.error("ERROR: GPS data is NOT available")
            return

        self._config_target_gps()

        target = None
        target_update = True
        takeoff_landing = self._takeoff_landing()
        if takeoff_landing:
            target = self._target_gps. \
                get_takeoff_landing_data(self._d_inDrCtrl.homeGPS, takeoff_landing)
        else:
            mv = self._moving()
            move_continue = self._move_continue(mv.front_back, mv.right_left, mv.up_down)
            if move_continue:
                dist_cal = drone_distance_cal.DroneMathmetics()
                dist = dist_cal.distance_math(self._d_outTarget.pos.longitude,
                                              self._d_inDrCtrl.droneGPS.longitude,
                                              self._d_outTarget.pos.latitude,
                                              self._d_inDrCtrl.droneGPS.latitude,
                                              0,
                                              0)

                if self._d_outMoveMode.data == "Absolute":
                    if mv.front_back:
                        update_coefficient = MOVE_UPDATE_COEF_ABSOLUTE_NS
                    else:
                        update_coefficient = MOVE_UPDATE_COEF_ABSOLUTE_EW
                elif self._d_outMoveMode.data == "Relative":
                    update_coefficient = MOVE_UPDATE_COEF_RELATIVE

                if self._last_speedparam == self._SpeedParam:
                    if (self._SpeedParam * self._LatLonCoefficient[0] * update_coefficient) <= dist:
                        target_update = False

            else:
                self._set_target_base_alt(mv)

            if target_update:
                target = self._target_gps.get_move_target_data(self._d_inDrCtrl.droneGPS,
                                                               mv.front_back,
                                                               mv.right_left,
                                                               mv.up_down,
                                                               self._SpeedParam,
                                                               self._d_inDrCtrl.mag.data)

            self._last_front_back = mv.front_back
            self._last_right_left = mv.right_left
            self._last_up_down = mv.up_down

        if target_update:
            self._d_outTarget.pos.latitude = target.latitude
            self._d_outTarget.pos.longitude = target.longitude
            self._d_outTarget.pos.altitude = target.altitude

    def _set_target_base_alt(self, mv):

        if self._last_up_down in [1, -1] and mv.up_down == 0:
            # Case: Alt-chg-req cancel(=The base of target data is the alt of drone)
            self._alt_chg_req_cancel = True
            self._alt_chg_cancel_wait_tm = 0
        else:
            # Case: Move-req, Move-req cancel and Hovering
            if self._alt_chg_req_cancel:
                self._alt_chg_cancel_wait_tm += 1
                if self._alt_chg_cancel_wait_tm == ALT_CHG_CANCEL_WAIT_TM:
                    # Update the base of target data w/ out-port target-alt
                    self._d_inDrCtrl.droneGPS.altitude = self._d_outTarget.pos.altitude
                    self._alt_chg_req_cancel = False
                    self._alt_chg_cancel_wait_tm = 0
            else:
                # The base of target data is the out-port target-alt
                self._d_inDrCtrl.droneGPS.altitude = self._d_outTarget.pos.altitude

    def _set_target_speed(self):
        self._move_speed()
        self._d_outTarget.speed.data = self._SpeedParam
        self._d_outTargetSpeed.data = self._SpeedParam

    def _set_target_yaw(self):
        target_deg = 0  
        target_dir = int(self._turn_dir())
        mv1 = self._front_back()
        mv2 = self._right_left()
        mv3 = self._up_down()

        move_continue = self._move_continue(mv1, mv2, mv3)
        if move_continue == False:
            self._relative_mag = int(self._d_inDrCtrl.mag.data)

        if target_dir:
            if mv1 or mv2 or mv3:
                turn_speed = self._MoveTurnSpeed[0]
            else:
                turn_speed = self._TurnSpeed[0]

            target_deg = int(self._d_inDrCtrl.mag.data) + (turn_speed * target_dir)
        else:
            if mv1 or mv2:
                if self._d_outMoveMode.data == "Absolute":
                    if self._last_front_back or self._last_right_left:
                        target_deg = int(self._diff_target_angle())
                        if target_deg > 0:
                            target_dir = 1
                        else:
                            target_dir = -1
                        target_deg += int(self._d_inDrCtrl.mag.data)

                elif self._d_outMoveMode.data == "Relative":
                    target_dir = 1
                    target_deg = self._relative_mag


        if target_deg > 360:
            target_deg = target_deg % 360
        elif target_deg < 0:
            target_deg = 360 - (abs(target_deg) % 360)

        self._d_outTarget.yaw.data = [target_dir,target_deg]

    def _set_move_mode(self):
        mode = self._move_mode()

        if mode == 1:
            self._d_outMoveMode.data = "Absolute"
        else:
            self._d_outMoveMode.data = "Relative"

        self._target_gps.set_controller_mode(mode)

    def _set_outport(self):
        
        self._out_target_wrote = False

        if self._drone_gps:
            if self._is_valid_target(self._d_outTarget, 0):
                self._d_outTarget.id += 1
                OpenRTM_aist.setTimestamp(self._d_outTarget)
                if self._outTargetOut.write():
                    self._out_target_wrote = True
                    logger.debug("TargetGPS: % f % f % f",
                            self._d_outTarget.pos.latitude,
                            self._d_outTarget.pos.longitude,
                            self._d_outTarget.pos.altitude)
                    logger.debug("TargetSpeed: %f",
                                self._d_outTarget.speed.data)
                    logger.debug("TargetYaw: (dir)%d, (deg)%d",
                                self._d_outTarget.yaw.data[0],
                                self._d_outTarget.yaw.data[1])
                    logger.debug("Send ID: % d",
                                self._d_outTarget.id)

        self._outTargetSpeedOut.write()
        OpenRTM_aist.setTimestamp(self._d_outMoveMode)
        self._outMoveModeOut.write()

        self._last_speedparam = self._SpeedParam
        logger.debug("MoveMode: %s",
                     self._d_outMoveMode.data)
        self._last_move_mode = self._d_outMoveMode.data

    def _diff_target_angle(self):
        angle_vec_x = math.cos(math.radians(self._d_inDrCtrl.mag.data))
        angle_vec_y = math.sin(math.radians(self._d_inDrCtrl.mag.data))

        target_vec_x = self._d_outTarget.pos.latitude - self._d_inDrCtrl.droneGPS.latitude
        target_vec_y = self._d_outTarget.pos.longitude - self._d_inDrCtrl.droneGPS.longitude

        inner_product = angle_vec_x * target_vec_x + angle_vec_y * target_vec_y
        cross_product = angle_vec_x * target_vec_y - angle_vec_y * target_vec_x

        angle_rad = math.atan2(cross_product, inner_product)
        angle_deg = math.degrees(angle_rad)

        return angle_deg

    def _check_drone_on_home(self):
        ret = False

        dist_cal = drone_distance_cal.DroneMathmetics()
        dist = dist_cal.distance_math(self._d_inDrCtrl.homeGPS.longitude,
                                      self._d_inDrCtrl.droneGPS.longitude,
                                      self._d_inDrCtrl.homeGPS.latitude,
                                      self._d_inDrCtrl.droneGPS.latitude,
                                      0,
                                      0)

        # TODO: 最適な誤差をシステムテストで割り出す
        if (dist <= ALLOWABLE_ERROR_POS_CMP and
                self._d_inDrCtrl.droneGPS.altitude < (self._MinAlt[0] - ALLOWABLE_ERROR_HEIGHT)):

            ret = True

        return ret

    def _takeoff_landing(self):
        operation = 0
        home = self._check_drone_on_home()
        if home:
            # check takeoff request
            if self._d_ButtonIn.data & BUTTON_06_MASK:
                operation = 1
            elif self._d_ButtonIn.data & BUTTON_08_MASK:
                logger.warning("WARNING: Landing operation is "
                               "NOT available "
                               "in the current drone position")
        else:
            # check landing request
            if self._d_ButtonIn.data & BUTTON_08_MASK:
                operation = -1
            elif self._d_ButtonIn.data & BUTTON_06_MASK:
                logger.warning("WARNING: Take-off operation is "
                               "NOT available in the current drone position")

        return operation

    def _move_mode(self):
        home = self._check_drone_on_home()
        mode = self._target_gps._mode
        
        if home:
            if self._d_ButtonIn.data & BUTTON_10_MASK:
                logger.warning("WARNING: MoveMode operation is "
                               "NOT available in the current drone position")                
        else:
            if self._d_ButtonIn.data & BUTTON_10_MASK:
                if mode == 1:
                    mode = 2
                else:
                    mode = 1

        return mode

    def _moving(self):
        mv1 = self._front_back()
        mv2 = self._right_left()
        mv3 = self._up_down()

        home = self._check_drone_on_home()

        if home:
            if mv1 or mv2 or mv3:
                mv1 = mv2 = mv3 = 0
                logger.warning("WARNING: Move operations are "
                               "NOT available in the current drone position")

        return self.MovingOperation(mv1, mv2, mv3)

    def _front_back(self):
        key = 0
        try:
            key = self._d_AnalogIn.data[0]
        except IndexError:
            pass

        operation = self._front_back_check(key)
        operation *= -1

        return operation

    def _right_left(self):
        operation = 0
        both_push = BUTTON_03_MASK | BUTTON_01_MASK
        if (self._d_ButtonIn.data & both_push) == both_push:
            logger.warning("WARNING: "
                           "Conflicting operations are detected [right/left]")
        elif self._d_ButtonIn.data & BUTTON_03_MASK:
            operation = 1
        elif self._d_ButtonIn.data & BUTTON_01_MASK:
            operation = -1

        return operation

    def _up_down(self):
        operation = 0
        both_push = BUTTON_04_MASK | BUTTON_02_MASK
        if (self._d_ButtonIn.data & both_push) == both_push:
            logger.warning("WARNING: "
                           "Conflicting operations are detected [up/down]")
        elif self._d_ButtonIn.data & BUTTON_04_MASK:
            operation = 1
        elif self._d_ButtonIn.data & BUTTON_02_MASK:
            operation = -1

        return operation

    def _move_continue(self, front_back, right_left, up_down):
        if self._last_move_mode == self._d_outMoveMode.data:
            if self._turn_dir():
                move_continue = False
            else:
                if (front_back != self._last_front_back) or (right_left != self._last_right_left) or (up_down != self._last_up_down):
                    move_continue = False
                elif (front_back == 0) and (right_left == 0) and (up_down == 0):
                    move_continue = False
                else:
                    move_continue = True
        else:
            move_continue = False

        return move_continue

    def _turn_dir(self):
        home = self._check_drone_on_home()
        key = 0

        try:
            key = self._d_AnalogIn.data[1]
        except IndexError:
            pass

        if home:
            if key:
                key = 0
                logger.warning("WARNING: Turning operation is "
                               "NOT available in the current drone position")

        else:
            if self._d_ButtonIn.data & BUTTON_08_MASK:
                key = 0
            else:
                if self._d_outMoveMode.data == "Absolute":
                    mv1 = self._front_back()
                    mv2 = self._right_left()
                    mv3 = self._up_down()
                    if mv1 or mv2 or mv3:
                        key = 0

        operation = self._turn_dir_check(key)

        return operation

    def _move_speed(self):
        home = self._check_drone_on_home()
        both_push = BUTTON_05_MASK | BUTTON_07_MASK
        
        if home:
            if (self._d_ButtonIn.data & BUTTON_05_MASK) or (self._d_ButtonIn.data & BUTTON_07_MASK) or (self._d_ButtonIn.data & both_push):
                logger.warning("WARNING: Speed operation is "
                               "NOT available in the current drone position") 
        elif (self._d_ButtonIn.data & both_push) == both_push:
            logger.warning("WARNING: "
                           "Conflicting operations are detected [MoveSpeed up/down]")
        elif self._d_ButtonIn.data & BUTTON_05_MASK:
            self._SpeedParam += 1.0

        elif self._d_ButtonIn.data & BUTTON_07_MASK:
            self._SpeedParam -= 1.0

        self._SpeedParam = self._move_speed_check(self._SpeedParam)

    def _front_back_check(self, status):
        if status > 1:
            status = 1
        elif status < -1:
            status = -1
            
        return status

    def _turn_dir_check(self, dir):
        if dir > 1:
            dir = 1
        elif dir < -1:
            dir = -1
            
        return dir

    def _move_speed_check(self, speed):
        if speed > MOVE_SPEED_MAX:
            speed = MOVE_SPEED_MAX
        elif speed < MOVE_SPEED_MIN:
            speed = MOVE_SPEED_MIN
            
        return speed
    

    def _cb_connect_status_change(self, port, connect):
        if port == "ButtonIn":
            self._button = connect
        elif port == "AnalogIn":
            self._analog = connect
        else:
            logger.warning("Invalid port name: %s", port)

        if not connect:
            logger.warning("ERROR: Data port was disconnected: %s", port)

    def _is_valid_drctrl(self, drctrl, num):
        ret = False

        Id = self._is_valid_id(drctrl.id, num, "Rcv ID")
        home_lat = self._is_valid_gps_lat(drctrl.homeGPS.latitude, num, "HomeGPS")
        home_lon = self._is_valid_gps_lon(drctrl.homeGPS.longitude, num, "HomeGPS")
        drone_lat = self._is_valid_gps_lat(drctrl.droneGPS.latitude, num, "DroneGPS")
        drone_lon = self._is_valid_gps_lon(drctrl.droneGPS.longitude, num, "DroneGPS")
        drone_alt = self._is_valid_gps_alt(drctrl.droneGPS.altitude, drctrl.state.data, num, "DroneGPS")
        mag = self._is_valid_angle(drctrl.mag.data, num, "Mag")
        state = self._is_valid_state(drctrl.state.data, num)

        if Id and home_lat and home_lon and drone_lat and drone_lon and drone_alt and mag and state:
            ret = True

        return ret

    def _is_valid_target(self, target, num):
        ret = False

        Id = self._is_valid_id(target.id, num, "Send ID")
        target_lat = self._is_valid_gps_lat(target.pos.latitude, num, "TargetGPS")
        target_lon = self._is_valid_gps_lon(target.pos.longitude, num, "TargetGPS")
        target_alt = self._is_valid_gps_alt(target.pos.altitude, self._d_inDrCtrl.state.data, num, "TargetGPS")
        target_speed = self._is_valid_speed(target.speed.data, num)
        target_yaw_dir = self._is_valid_angle_dir(target.yaw.data[0], num, "TargetYaw_dir")
        target_yaw_angle = self._is_valid_angle(target.yaw.data[1], num, "TargetYaw_angle")

        if Id and target_lat and target_lon and target_alt and target_speed and target_yaw_dir and target_yaw_angle:
            ret = True

        return ret

    def _is_valid_id(self, Id, num, name):
        ret = False
        if Id >= 0:
            ret = True
        else:
            logger.debug("invalid %s[%d]:  % d", name, num, Id)

        return ret

    def _is_valid_gps_lat(self, lat, num, name):
        ret = False
        if -90.0 <= lat <= 90.0:
            ret = True
        else:
            logger.debug("invalid %s[%d] lat:  % f", name, num, lat)

        return ret

    def _is_valid_gps_lon(self, lon, num, name):
        ret = False
        if -180.0 <= lon <= 180.0:
            ret = True
        else:
            logger.debug("invalid %s[%d] lon:  % f", name, num, lon)

        return ret

    def _is_valid_gps_alt(self, alt, state, num, name):
        ret = False
        if state in ["Hovering","In_Move","Alt_CHG","Yaw_CHG"]:
            if self._MinAlt[0] <= alt <= self._MaxAlt[0]:
                ret =  True
            else:
                if self._d_inDrCtrl.homeGPS.altitude == self._d_outTarget.pos.altitude:
                    # RTL
                    ret = True
                else:
                    logger.debug("invalid %s[%d] alt:  % f", name, num, alt)
        else:
            ret = True

        return ret

    def _is_valid_angle(self, angle, num, name):
        ret = False
        if 0 <= angle <= 360:
            ret = True
        else:
            logger.debug("invalid %s[%d]:  % d", name, num, angle)

        return ret

    def _is_valid_angle_dir(self, angle_dir, num, name):
        ret = False
        if angle_dir in [-1, 0, 1]:
            ret = True
        else:
            logger.debug("invalid %s[%d]:  % d", name, num, angle_dir)

        return ret

    def _is_valid_state(self, state, num):
        ret = False
        if state in ["Armed","Takeoff","Hovering","In_Move","Alt_CHG","Yaw_CHG","RTL","Disarm"]:
            ret = True
        else:
            logger.debug("invalid State[%d]:  % s", num, state)

        return ret

    def _is_valid_speed(self, speed, num):
        ret = False
        if speed >= 0:
            ret = True
        else:
            logger.debug("invalid TargetSpeed[%d]:  % f", num, speed)

        return ret

    def _ready_to_log_write(self, file_mode):
    
        if file_mode == FILE_MODE_WRITE:
            date, time = self._get_datetime_to_str()
            time = time.replace(":", "").replace(".", "")  #  HH:mm:ss.sss --> HHmmssSSS
            self._file_name = date + "_" + time[0:6] + "_ctrl_log.csv"

        if self._is_open_rtc_viewer_log(file_mode):
            if self._is_csv_writer():
                if file_mode == FILE_MODE_WRITE:
                    self._writer.writerow(self._header)
        
    def _get_datetime_to_str(self):
        
        curr_time = datetime.now()
        date = "%s%s%s" % ("{0:04d}".format(curr_time.year), "{0:02d}".format(curr_time.month), "{0:02d}".format(curr_time.day))
        time = "%s:%s:%s.%s" % ("{0:02d}".format(curr_time.hour), "{0:02d}".format(curr_time.minute),
                           "{0:02d}".format(curr_time.second), "{0:03d}".format(curr_time.microsecond / 1000))
        return date, time

    def _is_open_rtc_viewer_log(self, file_mode):

        rtn = False
        
        try:
            if self._file_name != "":
                self._file = open(self._file_name, file_mode)
                rtn = True
        except IOError as e:
            logger.error("I/O Error {0} {1}".format(e.errno, e.strerror))

        return rtn
                
    def _is_csv_writer(self):

        rtn = False

        try:
            if self._file is not None:
                self._writer = csv.writer(self._file)
                if self._writer is not None:
                    rtn = True
        except csv.Error as e:
            logger.error("CSV writer error {0} {1}".format(e.errno, e.strerror))

        return rtn

    def _write_rtc_viewer_log(self, rtc_state):

        if self._writer is not None:
            self._set_log(rtc_state)
            self._writer.writerow(self._log_row)
    
    def _close_rtc_viewer_log(self):

        if self._file is not None:
            self._file.close()
            self._file = None
            self._writer = None

    def _conv_rtctime_to_datetime(self, tm):

        curr_time = datetime.fromtimestamp(tm.sec)

        date = "%s%s%s" % ("{0:04d}".format(curr_time.year), "{0:02d}".format(curr_time.month),
                           "{0:02d}".format(curr_time.day))
        time = "%s:%s:%s.%s" % ("{0:02d}".format(curr_time.hour), "{0:02d}".format(curr_time.minute),
                                "{0:02d}".format(curr_time.second), "{0:03d}".format(tm.nsec / 1000000))

        return date, time

    def _set_log(self, rtc_state):
    
        self._set_rcv_log(rtc_state)
        self._set_snd_log(rtc_state)
        self._set_other_log(rtc_state)

    def _set_rcv_log(self, rtc_state):
 
        if rtc_state == "Active":
            if self._d_inDrCtrl.robotID.data:

                rcv_date, rcv_time = self._conv_rtctime_to_datetime(self._d_inDrCtrl.tm)

                self._log_row[self._header.index("RcvDate")] = rcv_date
                self._log_row[self._header.index("RcvTime")] = rcv_time
                self._log_row[self._header.index("RcvRobotID")] = self._d_inDrCtrl.robotID.data
                self._log_row[self._header.index("RcvID")] = self._d_inDrCtrl.id
                self._log_row[self._header.index("HomeLat")] = "{0:.10f}".format(self._d_inDrCtrl.homeGPS.latitude)
                self._log_row[self._header.index("HomeLon")] = "{0:.10f}".format(self._d_inDrCtrl.homeGPS.longitude)
                self._log_row[self._header.index("HomeAlt")] = "{0:.2f}".format(self._d_inDrCtrl.homeGPS.altitude)
                self._log_row[self._header.index("RobotLat")] = "{0:.10f}".format(self._d_inDrCtrl.droneGPS.latitude)
                self._log_row[self._header.index("RobotLon")] = "{0:.10f}".format(self._d_inDrCtrl.droneGPS.longitude)
                self._log_row[self._header.index("RobotAlt")] = "{0:.2f}".format(self._d_inDrCtrl.droneGPS.altitude)
                self._log_row[self._header.index("AccX")] = "{0:.2f}".format(self._d_inDrCtrl.accel.data[0])
                self._log_row[self._header.index("AccY")] = "{0:.2f}".format(self._d_inDrCtrl.accel.data[1])
                self._log_row[self._header.index("AccZ")] = "{0:.2f}".format(self._d_inDrCtrl.accel.data[2])
                self._log_row[self._header.index("Mag")] = int(self._d_inDrCtrl.mag.data)
                self._log_row[self._header.index("State")] = self._d_inDrCtrl.state.data

    def _set_snd_log(self, rtc_state):

        if rtc_state == "Active":
            snd_date, snd_time = self._conv_rtctime_to_datetime(self._d_outMoveMode.tm)

            self._log_row[self._header.index("SndDate")] = snd_date
            self._log_row[self._header.index("SndTime")] = snd_time
            self._log_row[self._header.index("MoveMode")] = self._d_outMoveMode.data

            if self._out_target_wrote:
                snd_date, snd_time = self._conv_rtctime_to_datetime(self._d_outTarget.tm)
                self._log_row[self._header.index("SndDate")] = snd_date
                self._log_row[self._header.index("SndTime")] = snd_time
                self._log_row[self._header.index("SndRobotID")] = self._d_inDrCtrl.robotID.data
                self._log_row[self._header.index("SndID")] = self._d_outTarget.id
                self._log_row[self._header.index("TargetLat")] = "{0:.10f}".format(self._d_outTarget.pos.latitude)
                self._log_row[self._header.index("TargetLon")] = "{0:.10f}".format(self._d_outTarget.pos.longitude)
                self._log_row[self._header.index("TargetAlt")] = "{0:.2f}".format(self._d_outTarget.pos.altitude)
                self._log_row[self._header.index("TargetSpd")] = "{0:.2f}".format(self._d_outTarget.speed.data)
                self._log_row[self._header.index("TargetYawDir")] = self._d_outTarget.yaw.data[0]
                self._log_row[self._header.index("TargetYawDeg")] = self._d_outTarget.yaw.data[1]

    def _set_other_log(self, rtc_state):

        self._log_row[self._header.index("Category")] = "GCS"
        self._log_row[self._header.index("ActStartDate")] = self._act_start_date
        self._log_row[self._header.index("ActStartTime")] = self._act_start_time
        self._log_row[self._header.index("RTCState")] = rtc_state


    class MovingOperation(object):

        def __init__(self, front_back, right_left, up_down):
            self.front_back = front_back
            self.right_left = right_left
            self.up_down = up_down


class ConnectionCallback(OpenRTM_aist.PortCallBack.ConnectionCallback):

    def __init__(self, func):
        self._plugin_func = func

    def __call__(self, profile):
        self._plugin_func()


def DroneControllerInit(manager):
    profile = OpenRTM_aist.Properties(defaults_str=dronecontroller_spec)
    manager.registerFactory(profile,
                            DroneController,
                            OpenRTM_aist.Delete)

def MyModuleInit(manager):
    DroneControllerInit(manager)

    # Create a component
    comp = manager.createComponent("DroneController")

def main():
    mgr = OpenRTM_aist.Manager.init(sys.argv)
    mgr.setModuleInitProc(MyModuleInit)
    mgr.activateManager()
    mgr.runManager()

if __name__ == "__main__":
    main()
