#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file target_gps.py
 @brief This module makes drone target GPS data.
 @date $Date$

 @author East Japan Accounting Center Co., Ltd.

 (c)2017 The University of Aizu
 This software is released under the New BSD License.

"""
import sys
import RTC
from logging import getLogger, StreamHandler
from logging import DEBUG as LOG_LV
import math

sys.path.append(".")

# constants
CONTROL_MODE_1 = 1     #Absolute
CONTROL_MODE_2 = 2     #Relative



# Setup logger
logger = getLogger(__name__)
handler = StreamHandler()
handler.setLevel(LOG_LV)
logger.setLevel(LOG_LV)
logger.addHandler(handler)


class TargetGPS(object):
    def __init__(self):
        self._init_variable()

    def _init_variable(self):
        self._mode = CONTROL_MODE_1
        self._AltCoefficient = 1
        self._LatLonCoefficient = 2
        self._LatDiff10m = 0.000137
        self._LonDiff10m = 0.000108
        self._MoveSpeed = 3.0
        self._MinAlt = 2.0
        self._MaxAlt = 15.0
        self._TurnSpeed = 45
        self._NoseLatDiff10m = 9.01323
        self._NoseLonDiff10m = 8.9831566
        self._MoveTurnSpeed = 3

    def config_param(self, kwargs):
        key1 = key2 = key3 = key4 = key5 = key6 = key7 = key8 = key9 = key10 = key11 = None
        try:
            key1 = kwargs["AltCoefficient"]
            key2 = kwargs["LatLonCoefficient"]
            key3 = kwargs["LatDiff10m"]
            key4 = kwargs["LonDiff10m"]
            key5 = kwargs["MoveSpeed"]
            key6 = kwargs["MinAlt"]
            key7 = kwargs["MaxAlt"]
            key8 = kwargs["TurnSpeed"]
            key9 = kwargs["NoseLatDiff10m"]
            key10 = kwargs["NoseLonDiff10m"]
            key11 = kwargs["MoveTurnSpeed"]

        except KeyError:
            logger.warning("ERROR: KeyError")

        if key1 is not None:
            self._AltCoefficient = key1

        if key2 is not None:
            self._LatLonCoefficient = key2

        if key3 is not None:
            self._LatDiff10m = key3

        if key4 is not None:
            self._LonDiff10m = key4

        if key5 is not None:
            self._MoveSpeed = key5

        if key6 is not None:
            self._MinAlt = key6

        if key7 is not None:
            self._MaxAlt = key7

        if key8 is not None:
            self._TurnSpeed = key8

        if key9 is not None:
            self._NoseLatDiff10m = key9

        if key10 is not None:
            self._NoseLonDiff10m = key10

        if key11 is not None:
            self._MoveTurnSpeed = key11

    def set_controller_mode(self, mode):
        if mode == CONTROL_MODE_1:
            self._mode = CONTROL_MODE_1
        else:
            self._mode = CONTROL_MODE_2

    def get_takeoff_landing_data(self, home, takeoff_landing):
        takeoff_landing = self._button_status_check(takeoff_landing)

        target = None
        if takeoff_landing:
            target = RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                 home.latitude, home.longitude, home.altitude,
                                 0.0, 0.0, 0.0, 0.0,
                                 0.0, 0, RTC.GPS_FIX_NORMAL)

            if takeoff_landing == 1:
                target.altitude = 10.0

        return target

    def get_move_target_data(self, drone, front_back, right_left, up_down, speedparam, mag):
        target = None
        if self._mode == CONTROL_MODE_1:
            target = self._move_mode1(drone, front_back, right_left, up_down, speedparam)
        else:
            target = self._move_mode2(drone, front_back, right_left, up_down, speedparam, mag)

        return target

    def _move_mode1(self, drone, south_north, east_west, up_down, speedparam):
        target = RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0, RTC.GPS_FIX_NORMAL)

        target.latitude = self._move_south_north(drone.latitude, south_north, speedparam)
        target.longitude = self._move_east_west(drone.longitude, east_west, speedparam)
        target.altitude = self._move_up_down(drone.altitude, up_down)

        return target

    def _move_mode2(self, drone, front_back, right_left, up_down, speedparam, mag):
        target = RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0, RTC.GPS_FIX_NORMAL)

        target.latitude,target.longitude = self._move_nose(drone, front_back, right_left, speedparam, mag)
        target.altitude = self._move_up_down(drone.altitude, up_down)

        return target

    def _move_south_north(self, drone_lat, operation, speedparam):
        operation = self._analog_status_check(operation)

        diff = ((self._LatDiff10m / 10) *
                speedparam *
                self._LatLonCoefficient *
                operation)

        target_lat = drone_lat + diff

        if operation:
            target_lat = self._lat_range_check(target_lat)

        return target_lat

    def _move_east_west(self, drone_lon, operation, speedparam):
        operation = self._button_status_check(operation)

        diff = ((self._LonDiff10m / 10) *
                speedparam *
                self._LatLonCoefficient *
                operation)

        target_lon = drone_lon + diff

        if operation:
            target_lon = self._lon_range_check(target_lon)

        return target_lon

    def _move_up_down(self, drone_alt, operation):
        operation = self._button_status_check(operation)

        diff = self._AltCoefficient * operation
        target_alt = drone_alt + diff

        if operation:
            target_alt = self._alt_range_check(target_alt)

        return target_alt

    def _move_nose(self, drone, front_back, right_left, speedparam, mag):
        operation = 0.0

        front_back = self._analog_status_check(front_back)
        right_left = self._button_status_check(right_left)

        if (front_back == 1 and right_left == 1) or (front_back == -1 and right_left == -1):
            mag += 45.0
            operation = front_back
        elif (front_back == -1 and right_left == 1) or (front_back == 1 and right_left == -1):
            mag += 135.0
            operation = right_left
        elif front_back:
            operation = front_back
        elif right_left:
            mag += 90.0
            operation = right_left
            
        if mag > 360.0:
            mag = mag % 360.0
        elif mag < 0.0:
            mag = 360.0 - (abs(mag) % 360.0)


        latdiff = ((self._NoseLatDiff10m / 10) *
                pow(10, -5) *
                speedparam *
                self._LatLonCoefficient *
                math.cos(math.radians(mag)) *
                operation)
        londiff = (((self._NoseLonDiff10m / 10) *
                pow(10, -5) /
                math.cos(latdiff / (2 * math.pi))) *
                speedparam *
                self._LatLonCoefficient *
                math.sin(math.radians(mag)) *
                operation)

        target_lat = drone.latitude + latdiff
        target_lon = drone.longitude + londiff

        if operation:
            target_lat = self._lat_range_check(target_lat)             
            target_lon = self._lon_range_check(target_lon)

        return [target_lat, target_lon]

    def _button_status_check(self, button):
        if button > 1:
            button = 1
        elif button < -1:
            button = -1
            
        return button

    def _analog_status_check(self, analog):
        if analog > 1:
            analog = 1
        elif analog < -1:
            analog = -1
            
        return analog

    def _alt_range_check(self, alt):
        if alt > self._MaxAlt:
            alt = self._MaxAlt
        elif alt < self._MinAlt:
            alt = self._MinAlt
            
        return alt

    def _lat_range_check(self, lat):
        if lat > 90.0:
            lat = 90.0
        elif lat < -90.0:
            lat = -90.0
        
        return lat

    def _lon_range_check(self, lon):
        if lon > 180.0:
            lonRem = lon % 360.0
            if lonRem > 180.0:
                lon = lonRem - 360.0
            else:
                lon = lonRem
        elif lon < -180.0:
            lonRem = abs(lon) % 360.0
            if lonRem > 180.0:
                lon = 360.0 - lonRem
            else:
                lon = -lonRem

        return lon


def debug_target_gps_data():
    m = TargetGPS()
    config = {"AltCoefficient": 111,
              "LatLonCoefficient": 222,
              "LatDiff10m": 333.333,
              "LonDiff10m": 444.444,
              "MoveSpeed": 555.5555,
              "MinAlt": 6.0,
              "MaxAlt": 77.0,
              "TurnSpeed": 88,
              "NoseLatDiff10m": 999.999,
              "NoseLonDiff10m": 1001.1001,
              "MoveTurnSpeed": 1111
              }

    m.config_param(config)
    logger.debug("%d %d %f %f %f\n" 
                 "%f %f %d %f %f %d",
                 m._AltCoefficient, m._LatLonCoefficient,
                 m._LatDiff10m, m._LonDiff10m, m._MoveSpeed,
                 m._MinAlt, m._MaxAlt, m._TurnSpeed,
                 m._NoseLatDiff10m, m._NoseLonDiff10m, m._MoveTurnSpeed)

    m._init_variable()
    gps_data = RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                           37.523763, 139.938025, 3.5, 0.0, 0.0, 0.0, 0.0,
                           0.0, 0, RTC.GPS_FIX_NORMAL)

    m.set_controller_mode(2)
    logger.debug(m.get_takeoff_landing_data(gps_data, 1))
    logger.debug(m.get_takeoff_landing_data(gps_data, 0))
    logger.debug(m.get_takeoff_landing_data(gps_data, -1))

    logger.debug(m.get_move_target_data(gps_data, -1, 1, 0, 3.0, 405.0))
    logger.debug(m.get_move_target_data(gps_data, 1, -1, 0, 3.0, 405.0))
    logger.debug(m.get_move_target_data(gps_data, 1, 0, 0, 3.0, -675.0))

if __name__ == "__main__":
    debug_target_gps_data()
